"""StegoGAN example: non-bijective unpaired image translation training."""

from examples.stegogan.train_stegogan import StegoGANConfig, StegoGANTrainer

__all__ = [
    "StegoGANTrainer",
    "StegoGANConfig",
]
