"""StarGAN example: multi-domain image translation training."""

from examples.stargan.train_stargan import StarGANTrainer, StarGANTrainingConfig

__all__ = [
    "StarGANTrainer",
    "StarGANTrainingConfig",
]
