# Copyright (c) 2026 EarthBridge Team.
# Credits: OpenEarthMap-SAR CUT models. Architecture from Park et al. "Contrastive Learning for Unpaired Image-to-Image Translation" ECCV 2020.

"""Training stub for OpenEarthMap-SAR.

OpenEarthMap-SAR checkpoints are pre-trained. For training new CUT models
with the same architecture, use the main CUT training in
``examples/cut/train_cut.py`` with a generator compatible with
:class:`OpenEarthMapSARGenerator` (anti-aliased down/upsampling).
"""
