"""Pix2Pix example: paired image-to-image translation training."""

from examples.pix2pix.train_pix2pix import Pix2PixTrainer, TrainingConfig

__all__ = [
    "Pix2PixTrainer",
    "TrainingConfig",
]
